% *************************************************************************
% ********************************** OWE **********************************
% An On-line Weighted Ensemble of Regressor Models to Handle Concept Drifts
% *************************************************************************
% Version 1.0 Nov/08/2014
%
% Developed in MATLAB R2010a
% 
% Author and programmer: Symone Gomes Soares
%
% e-mail: symonesoares@gmail.com or symonesoares@isr.uc.pt
%
% Homepage: http://sites.google.com/site/symonesoares/
%
% Paper: SOARES, Symone Gomes; ARAÚJO, Rui. An On-line Weighted Ensemble 
%        of Regressor Models to Handle Concept Drifts. Engineering  
%        Applications of Artificial Intelligence, v. 37, p. 392-406, 2015. 
%     url: http://www.sciencedirect.com/science/article/pii/S09521976140024
%
% *************************************************************************
% *************************************************************************
% Inputs of the Algorithm:
%
% - input(number of samples x number of input variables) : input matrix
% 
% - output(number of samples x output variable) : output matrix 
%   ps.: it accepts only one output variable
%
% - m       : window's size (m >= 10)
% - Kappa   : discounting factor (0 <= Kappa < 1); Default: 0.2;
% - Theta   : for demarcating correct and incorrect prediction (0 < Theta < 1);
%             Default: 0.05;      
% - Alpha   : factor to include a new model (0 < Alpha < 1);
%             Default: 0.04; 
% - Pruning : ensemble pruning activation; introduce 1 for pruning
% - B       : maximum number of models in the ensemble, if Pruning = 1.
% 
% *************************************************************************
% Outputs of the Algorithm:
%
% CC   : correlation coefficient between predicted outputs and real outputs
% MSE  : mean square errord between predicted outputs and real outputs
% RMSE : root mean square errord between predicted outputs and real outputs
% nM   : number of models of the ensemble
% YR   : real outputs
% YP   : predicted outputs
% loss : loss between predicted outputs and real outputs
% time : running time of the algorithm
%
% *************************************************************************
% *************************************************************************
% *************************************************************************

function [CC, MSE, RMSE, nM, YR, YP, loss, time] = fOWE(input, output, m, Kappa, Theta, Alpha, B, Pruning)

    tStart = tic; 
    %Build the initial model
    n                 = 1; % parameter for the AdaBoost.RT 
    nS                = size(input,1); %Number of Samples
    nM                = 1;             %Setting Number of Models
    XShifted          = input(1:m,:);  %Assigning the inputs for the initial model
    YShifted          = output(1:m,:); %Assigning the outputs for the initial model
    [~, ~, T_press]   = KfoldCV(XShifted,YShifted,10); %Performing 10-fold cross-validation
    [~, Model(nM).lt] = min(T_press);               %Mininum PRESS and better number of latent variables for building the intial model 

    [mcxcal,Model(nM).mxcal] = mncn(XShifted);                  %Mean center for the X data
    [mcycal,Model(nM).mycal] = mncn(YShifted);                  %Mean center for the Y data
    Model(nM).bpls           = simpls(mcxcal,mcycal,Model(nM).lt); %Obtaining the initial model

    %Obtain the error of the initial model (nM=1)
    Yp             = rescale(mcxcal*Model(nM).bpls(Model(nM).lt,:)',Model(nM).mycal); %Predictions using the initial model using the X data
    AREt           = abs((Yp - YShifted) ./ YShifted); %Calculating the absolute relative error for the X data
    IX             = AREt > Theta;                     %Finding samples with error superior to the threshold l
    DEns           = ones(1,m) ./ m;                   %Normalizing the distribution
    Model(nM).e    = sum(DEns(1,IX)) + 0.00001;        %Calculating the error rate
    Model(nM).b    = Model(nM).e ^ n;                  %Setting B
    Model(nM).B    = Model(nM).b;
    Model(nM).AREk = AREt;

    %Some settings...
    nM   = nM + 1;  %...
    YP = []; % vector with the predicted outputs
    YR = []; % vector with the real outputs
    
    % on-line predictions
    for i = (m + 1):nS

        %Receiving a new sample
        xi = input(i,:);
        yi = output(i,:);
        
        % Making predictions;
        [yiEns, yiModels] = EnsemblePLSAdaBoostLog(Model, xi);
        %fprintf('\nSample [%d], real output: %f, predicted output: %f \n', i, yi, yiEns);
        YP = [YP yiEns];
        YR = [YR yi];

        %Sliding the window
        XShifted(1,:) = [];  
        YShifted(1,:) = [];
        XShifted = [XShifted; xi]; 
        YShifted = [YShifted; yi];

        %obtaining the error of the model on the new sample
        if (yi ~= 0)
           AREi = abs((yiEns - yi)/yi);
        else
           AREi = abs(yiEns - yi);
        end
        AREt(1,:) = [];
        AREt      = [AREt; AREi];

        %Obtain the new distribution...
        DEns  = ones(1,m) ./ m; %normalize
        Error = size(find(AREt > Theta),1);
        if (Error ~= 0)
          Upfactor   = m / Error;
          Downfactor = 1 / Upfactor;
          for j = 1:m
              if AREt(j,1) > Theta
                  DEns(1,j) = DEns(1,j) * Upfactor;
              else
                  DEns(1,j) = DEns(1,j) * Downfactor;
              end
          end
          DEns = DEns ./ sum(DEns);
        end 

        %Update AREk for the other models....
        for j = 1:size(Model,2)
           fEk = yiModels(1,j);
           if (yi ~= 0)
              AREik      = abs((fEk - yi)/yi);
              else AREik = abs(fEk - yi);
           end
           Model(j).AREk(1,:) = [];
           Model(j).AREk      = [Model(j).AREk; AREik];
        end 

        %Checking if a new model should be included
        if (AREi > Alpha)
           %fprintf('Creating a new model.... [Ensemble size = %d] \n', size(Model,2));
           
           %Generating a new model: Model(nM)
           [~, ~, T_press]          = KfoldCV(XShifted,YShifted, 10); %10-fold cross-validation
           [~, Model(nM).lt]        = min(T_press);
           [mcxcal,Model(nM).mxcal] = mncn(XShifted);                  %Mean center for the X data
           [mcycal,Model(nM).mycal] = mncn(YShifted);                  %Mean center for the Y data
           Model(nM).bpls           = simpls(mcxcal,mcycal,Model(nM).lt); %Obtaining Model(nM)
           Yp                       = rescale(mcxcal*Model(nM).bpls(Model(nM).lt,:)',Model(nM).mycal); %Predictions using the initial model using the X data
           Model(nM).AREk           = abs((Yp - YShifted) ./ YShifted); %Calculating the absolute relative error for the X data
           Model(nM).b              = [];
           nM                       = nM + 1;
        end

        %Evaluating all models on distribution DEns
        for j = 1:size(Model,2)
           IX         = Model(j).AREk > Theta;
           Model(j).e = sum(DEns(1,IX)) + 0.00001;
           Model(j).b = [Model(j).b Model(j).e ^ n];
           vBetas     = fReturnVecBeta(size(Model(j).b,2), Kappa);
           Model(j).B = (vBetas * Model(j).b') / sum(vBetas);
        end
        
        if size(Model,2) > B && Pruning == 1
              %fprintf('Excluding a model\n');
              maxError = -1;
              for j = 1:size(Model,2)-1
                if(Model(j).B >= maxError)
                   maxError = Model(j).B;
                   ix = j;
                end
              end
              Model(ix) = [];
              nM = nM - 1;
        end

    end
    
    time = toc(tStart);
    CC   = corr2(YR,YP);
    MSE  = mse(YP - YR);
    RMSE = mse(YP - YR)^1/2;
    nM   = size(Model,2);
    loss = (YP - YR).^2;
    fprintf('\n\n\nCorrelation coefficient: %f \n', CC);
    fprintf('Mean Squared Error:      %f \n', MSE);
    fprintf('Root Mean Squared Error: %f \n', RMSE);
    fprintf('Total number of models:  %d \n', nM);

end
